import numpy as np

def find_l(p, list_l, adj,sep=0.001, tol=0.01):
    print('...Finding optimal l for ','p=',p, '...',sep='')
    p_list = []
    for l in list_l:
        adj_exp=np.exp(-1*adj/(2*(l**2)))
        #print("l is ",str(l),"Percentage of total expression contributed by neighborhoods:",np.mean(np.sum(adj_exp,1))-1)
        p_list.append(np.mean(np.sum(adj_exp,1))-1)
    l1 = list_l[(np.abs(np.array(p_list)-p)).argmin()]
    if p_list[(np.abs(np.array(p_list)-0.5)).argmin()]<p:
        start = l1
        end = l1+0.1
    else:
        start = l1-0.1
        end = l1
    for i in np.arange(start, end, sep):
        adj_exp=np.exp(-1*adj/(2*(i**2)))
        q=np.mean(np.sum(adj_exp,1))-1
        #print("l=", str(i), "p=", str(round(q,5)))
        if np.abs(p-q)<tol:
            print("   l=", str(i), ", p=", str(round(q,5)), sep='')
            return i
    print("l not found, try bigger range or smaller sep!")





from keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau,History
from keras.layers import Input, Dense, Dropout
from keras.models import Model, Sequential
from keras.optimizers import SGD
from keras.utils.vis_utils import plot_model
import math


class SAE(object):
    """ 
    Stacked autoencoders. It can be trained in layer-wise manner followed by end-to-end fine-tuning.
    For a 5-layer (including input layer) example:
        Autoendoers model: Input -> encoder_0->act -> encoder_1 -> decoder_1->act -> decoder_0;
        stack_0 model: Input->dropout -> encoder_0->act->dropout -> decoder_0;
        stack_1 model: encoder_0->act->dropout -> encoder_1->dropout -> decoder_1->act;
    
    Usage:
        from SAE import SAE
        sae = SAE(dims=[784, 500, 10])  # define a SAE with 5 layers
        sae.fit(x, epochs=100)
        features = sae.extract_feature(x)
        
    Arguments:
        dims: list of number of units in each layer of encoder. dims[0] is input dim, dims[-1] is units in hidden layer.
              The decoder is symmetric with encoder. So number of layers of the auto-encoder is 2*len(dims)-1
        act: activation (default='relu'), not applied to Input, Hidden and Output layers.
        drop_rate: drop ratio of Dropout for constructing denoising autoencoder 'stack_i' during layer-wise pretraining
        batch_size: batch size
    """
    def __init__(self, dims, act='relu', drop_rate=0.2, batch_size=32,actinlayer1="tanh",init="glorot_uniform"): #act relu
        self.dims = dims
        self.n_stacks = len(dims) - 1
        self.n_layers = 2*self.n_stacks  # exclude input layer
        self.activation = act
        #self.actinlayer1="tanh" #linear
        self.actinlayer1=actinlayer1 #linear
        self.drop_rate = drop_rate
        self.init=init
        self.batch_size = batch_size
        self.stacks = [self.make_stack(i) for i in range(self.n_stacks)]
        self.autoencoders ,self.encoder= self.make_autoencoders()
        #plot_model(self.autoencoders, show_shapes=True, to_file='autoencoders.png')

    def make_autoencoders(self):
        """ Fully connected autoencoders model, symmetric.
        """
        # input
        x = Input(shape=(self.dims[0],), name='input')
        h = x

        # internal layers in encoder
        for i in range(self.n_stacks-1):
            h = Dense(self.dims[i + 1], kernel_initializer=self.init,activation=self.activation, name='encoder_%d' % i)(h)

        # hidden layer,default activation is linear
        h = Dense(self.dims[-1],kernel_initializer=self.init, name='encoder_%d' % (self.n_stacks - 1),activation=self.actinlayer1)(h)  # features are extracted from here

        y=h
        # internal layers in decoder       
        for i in range(self.n_stacks-1, 0, -1): #2,1
            y = Dense(self.dims[i], kernel_initializer=self.init,activation=self.activation, name='decoder_%d' % i)(y)

        # output
        y = Dense(self.dims[0], kernel_initializer=self.init,name='decoder_0',activation=self.actinlayer1)(y)

        return Model(inputs=x, outputs=y,name="AE"),Model(inputs=x,outputs=h,name="encoder")

    def make_stack(self, ith):
        """ 
        Make the ith denoising autoencoder for layer-wise pretraining. It has single hidden layer. The input data is 
        corrupted by Dropout(drop_rate)
        
        Arguments:
            ith: int, in [0, self.n_stacks)
        """
        in_out_dim = self.dims[ith]
        hidden_dim = self.dims[ith+1]
        output_act = self.activation
        hidden_act = self.activation
        if ith == 0:
            output_act = self.actinlayer1# tanh, or linear
        if ith == self.n_stacks-1:
            hidden_act = self.actinlayer1 #tanh, or linear
        model = Sequential()
        model.add(Dropout(self.drop_rate, input_shape=(in_out_dim,)))
        model.add(Dense(units=hidden_dim, activation=hidden_act, name='encoder_%d' % ith))
        model.add(Dropout(self.drop_rate))
        model.add(Dense(units=in_out_dim, activation=output_act, name='decoder_%d' % ith))

        #plot_model(model, to_file='stack_%d.png' % ith, show_shapes=True)
        return model

    def pretrain_stacks(self, x, epochs=200):
        """ 
        Layer-wise pretraining. Each stack is trained for 'epochs' epochs using SGD with learning rate decaying 10
        times every 'ep   ochs/3' epochs.
        
        Arguments:
            x: input data, shape=(n_samples, n_dims)
            epochs: epochs for each stack
        """
        #print("Doing SAE: pretrain_stacks")  
        features = x
        for i in range(self.n_stacks):
            print( '   ...Pretraining the %dth layer...' % (i+1))
            for j in range(3):  # learning rate multiplies 0.1 every 'epochs/3' epochs
                #print ('learning rate =', pow(10, -1-j))
                self.stacks[i].compile(optimizer=SGD(pow(10, -1-j), momentum=0.9), loss='mse')
                #callbacks=[EarlyStopping(monitor='loss',min_delta=10e-4,patience=4,verbose=1,mode='auto')]
                #self.stacks[i].fit(features, features, batch_size=self.batch_size, callbacks=callbacks,epochs=math.ceil(epochs/3))
                self.stacks[i].fit(features, features, batch_size=self.batch_size,epochs=math.ceil(epochs/3),verbose=0)
            print ('      The %dth layer has been pretrained' % (i+1))

            # update features to the inputs of the next layer
            feature_model = Model(inputs=self.stacks[i].input, outputs=self.stacks[i].get_layer('encoder_%d'%i).output)
            features = feature_model.predict(features)

    def pretrain_autoencoders(self, x, epochs=500):
        """
        Fine tune autoendoers end-to-end after layer-wise pretraining using 'pretrain_stacks()'
        Use SGD with learning rate = 0.1, decayed 10 times every 80 epochs
        
        :param x: input data, shape=(n_samples, n_dims)
        :param epochs: training epochs
        :return: 
        """
        #print("Doing SAE: pretrain_autoencoders")  
        print ('   ...Copying layer-wise pretrained weights to deep autoencoders...')
        for i in range(self.n_stacks):
            name = 'encoder_%d' % i
            self.autoencoders.get_layer(name).set_weights(self.stacks[i].get_layer(name).get_weights())
            name = 'decoder_%d' % i
            self.autoencoders.get_layer(name).set_weights(self.stacks[i].get_layer(name).get_weights())

        print ('   ...Fine-tuning autoencoder end-to-end...')
        for j in range(math.ceil(epochs/50)):
            lr = 0.1*pow(10, -j)
            #print ('learning rate =', lr)
            self.autoencoders.compile(optimizer=SGD(lr, momentum=0.9), loss='mse')
            #callbacks=[EarlyStopping(monitor='loss',min_delta=10e-4,patience=4,verbose=1,mode='auto')]
            #self.autoencoders.fit(x=x, y=x, batch_size=self.batch_size, epochs=50,callbacks=callbacks)
            self.autoencoders.fit(x=x, y=x, batch_size=self.batch_size, epochs=50,verbose=0)

    def fit(self, x, epochs=200):
        self.pretrain_stacks(x, epochs=epochs/2)
        self.pretrain_autoencoders(x, epochs=epochs)#fine tunning

    def fit2(self,x,epochs=200): #no stack directly traning 
        for j in range(math.ceil(epochs/50)):
            lr = 0.1*pow(10, -j)
            print ('learning rate =', lr)
            self.autoencoders.compile(optimizer=SGD(lr, momentum=0.9), loss='mse')
            self.autoencoders.fit(x=x, y=x, batch_size=self.batch_size, epochs=50)

    def extract_feature(self, x):
        """
        Extract features from the middle layer of autoencoders.
        
        :param x: data
        :return: features
        """
        hidden_layer = self.autoencoders.get_layer(name='encoder_%d' % (self.n_stacks - 1))
        feature_model = Model(self.autoencoders.input, hidden_layer.output)
        return feature_model.predict(x, batch_size=self.batch_size)


if __name__ == "__main__":
    """
    An example for how to use SAE model on MNIST dataset. In terminal run
            python SAE.py
    to see the result. You may get NMI=0.77.
    """
    import numpy as np

    def load_mnist():
        # the data, shuffled and split between train and test sets
        from keras.datasets import mnist
        (x_train, y_train), (x_test, y_test) = mnist.load_data()
        x = np.concatenate((x_train, x_test))
        y = np.concatenate((y_train, y_test))
        x = x.reshape((x.shape[0], -1))
        x = np.divide(x, 50.)  # normalize as it does in DEC paper
        print ('MNIST samples', x.shape)
        return x, y

    db = 'mnist'
    n_clusters = 10
    x, y = load_mnist()

    # define and train SAE model
    sae = SAE(dims=[x.shape[-1], 500, 500, 2000, 10])
    sae.fit(x=x, epochs=400)
    sae.autoencoders.save_weights('weights_%s.h5' % db)

    # extract features
    print ('Finished training, extracting features using the trained SAE model')
    features = sae.extract_feature(x)

    print ('performing k-means clustering on the extracted features')
    from sklearn.cluster import KMeans
    km = KMeans(n_clusters, n_init=20)
    y_pred = km.fit_predict(features)

    from sklearn.metrics import normalized_mutual_info_score as nmi
    print ('K-means clustering result on extracted features: NMI =', nmi(y, y_pred))



import scanpy as sc
import pandas as pd
import numpy as np
import scipy
import os
from anndata import AnnData,read_csv,read_text,read_mtx
from scipy.sparse import issparse
from natsort import natsorted
from anndata import read_mtx
from anndata.utils import make_index_unique


def read_10X(data_path, var_names='gene_symbols'):
    adata = read_mtx(data_path + '/matrix.mtx').T
    genes = pd.read_csv(data_path + '/genes.tsv', header=None, sep='\t')
    adata.var['gene_ids'] = genes[0].values
    adata.var['gene_symbols'] = genes[1].values
    assert var_names == 'gene_symbols' or var_names == 'gene_ids', \
        'var_names must be "gene_symbols" or "gene_ids"'
    if var_names == 'gene_symbols':
        var_names = genes[1]
    else:
        var_names = genes[0]
    if not var_names.is_unique:
        var_names = make_index_unique(pd.Index(var_names))
        print('var_names are not unique, "make_index_unique" has applied')
    adata.var_names = var_names
    cells = pd.read_csv(data_path + '/barcodes.tsv', header=None, sep='\t')
    adata.obs['barcode'] = cells[0].values
    adata.obs_names = cells[0]
    return adata



def change_to_continuous(q):
    #y_trans=q.argmax(axis=1)
    y_pred=np.asarray(np.argmax(q,axis=1),dtype=int)
    unique_labels=np.unique(q.argmax(axis=1))
    #turn to continuous clusters label,from 0,1,2,3,...
    test_c={}
    for ind, i in enumerate(unique_labels):
        test_c[i]=ind
    y_pred=np.asarray([test_c[i] for i in y_pred],dtype=int)
    ##turn to categories
    labels=y_pred.astype('U')
    labels=pd.Categorical(values=labels,categories=natsorted(np.unique(y_pred).astype('U')))
    return labels

def presettings(save_dir="result_scRNA",dpi=200,verbosity=3):
    if not os.path.exists(save_dir):
        print("Warning:"+str(save_dir)+"does not exists, so we  will creat it automatically!!:\n")
        os.mkdir(save_dir)
    figure_dir=os.path.join(save_dir,"figures")
    if not os.path.exists(figure_dir):
        os.mkdir(figure_dir)
    sc.settings.figdir=figure_dir
    sc.settings.verbosity=verbosity
    sc.settings.set_figure_params(dpi=dpi)
    sc.logging.print_versions()

def creatadata(datadir=None,exprmatrix=None,expermatrix_filename="matrix.mtx",is_mtx=True,cell_info=None,cell_info_filename="barcodes.tsv",gene_info=None,gene_info_filename="genes.tsv",project_name=None):
    """
    Construct a anndata object
    
    Construct a anndata from data in memory or files on disk. If datadir is a dir, there must be at least include "matrix.mtx" or data.txt(without anly columns name or rowname and sep="\t") , 

    """
    if (datadir is None and expermatrix is None and expermatrix_filename is None):
        raise ValueError("Please provide either the expression matrix or the ful path to the expression  matrix!!")
        #something wrong
    cell_info=pd.DataFrame(["cell_"+str(i) for i in range(1,x.shape[0]+1)],columns=["cellname"]) if cell_info is not None else cell_info
    gene_info=pd.DataFrame(["gene_"+str(i) for i in range(1,x.shape[1]+1)],columns=["genename"]) if gene_info is not None else gene_info 
    if datadir is not None:
        cell_and_gene_file = [f for f in os.listdir(datadir) if os.path.isfile(os.path.join(datadir, f))]
        if (os.path.isdir(datadir) and is_mtx==True): #sparse
            print("Start to read expression data (matrix.mtx)")
            x=sc.read_mtx(os.path.join(datadir,expermatrix_filename)).X.T
        else: #nonsparse
            x=pd.read_csv(os.path.join(datadir,expermatrix_filename),sep="\t",header=F)
       
            #only matrix with row names and colnames
        if cell_info_filename in cell_and_gene_file:
            cell_info=pd.read_csv(os.path.join(datadir,cell_info_filename),sep="\t",header=0,na_filter=False) 
        if gene_info_filename in cell_and_gene_file:
            gene_info=pd.read_csv(os.path.join(datadir,gene_info_filename),sep="\t",header=0,na_filter=False)
    else:
        x=exprmatrix # n*p matrix, cell* gene
  
    adata=sc.AnnData(x,obs=cell_info,var=gene_info)
    a=adata.obs["cellname"] if "cellname" in adata.obs.keys() else adata.obs.index
    adata.var_names=adata.var["genename"] if "genename" in adata.var.keys() else adata.var.index
    adata.obs_names_make_unique(join="-")
    adata.var_names_make_unique(join="-")
    adata.uns["ProjectName"]="DEC_clust_algorithm" if project_name is None else project_name 
    return adata

def prefilter_cells(adata,min_counts=None,max_counts=None,min_genes=200,max_genes=None):
    if min_genes is None and min_counts is None and max_genes is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[0],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_genes=min_genes)[0]) if min_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_genes=max_genes)[0]) if max_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_obs(id_tmp)
    adata.raw=sc.pp.log1p(adata,copy=True) #check the rowname 
    print("the var_names of adata.raw: adata.raw.var_names.is_unique=:",adata.raw.var_names.is_unique)
        
def prefilter_genes(adata,min_counts=None,max_counts=None,min_cells=10,max_cells=None):
    if min_cells is None and min_counts is None and max_cells is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[1],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_cells=min_cells)[0]) if min_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_cells=max_cells)[0]) if max_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_var(id_tmp)

def prefilter_specialgenes(adata,Gene1Pattern="ERCC",Gene2Pattern="MT-"):
    id_tmp1=np.asarray([not str(name).startswith(Gene1Pattern) for name in adata.var_names],dtype=bool)
    id_tmp2=np.asarray([not str(name).startswith(Gene2Pattern) for name in adata.var_names],dtype=bool)
    id_tmp=np.logical_and(id_tmp1,id_tmp2)
    adata._inplace_subset_var(id_tmp)
 
def normalize_log1p_scale(adata,units="UMI",n_top_genes=1000):
    if units=="UMI" or units== "CPM":
        sc.pp.normalize_per_cell(adata,counts_per_cell_after=10e4)
    sc.pp.filter_genes_dispersion(adata,n_top_genes=n_top_genes)
    sc.pp.log1p(adata)
    sc.pp.scale(adata,zero_center=True,max_value=6)
                    
#creat DEC object
def get_xinput(adata):
    if not isinstance(adata,AnnData):
        raise ValueError("adata must be an AnnData object")
    if issparse(adata.X):
        x=adata.X.toarray()
    else:
        x=adata.X
    return x

def getdims(dim):
    x=dim
    assert len(x)==2
    n_sample=x[0]
    if n_sample>20000:
        dims=[x[1],128,32]
    elif n_sample>10000:
        dims=[x[1],64,32]
    elif n_sample>5000:
        dims=[x[1],32,16]
    elif n_sample>2000:
        dims=[x[1],128]
    elif n_sample>500:
        dims=[x[1],64] #
    else:
        dims=[x[1],16] # or 32
    return dims

def OriginalClustering(adata,resolution=1.2,n_neighbors=20,n_comps=50,n_PC=20,n_job=4,dotsne=True,doumap=True,dolouvain=True):
    #Do PCA directly
    sc.tl.pca(adata,n_comps=n_comps)
    n_pcs=n_PC if n_PC<n_comps else n_comps
    #Do tsne based pca result
    if dotsne:
        sc.tl.tsne(adata,random_state=2,learning_rate=150,n_pcs=n_PC,n_jobs=n_job)
        #Save original X
        adata.obsm["X_tsne.ori"]=adata.obsm['X_tsne']
    #Do umap 
    if doumap:
        sc.pp.neighbors(adata,n_neighbors=n_neighbors)
        sc.tl.umap(adata)
        #Save original
        adata.obsm['X_umap.ori']=adata.obsm['X_umap']
    if dolouvain:
        sc.tl.louvain(adata,resolution=resolution)
        adata.obs['louvain_ori']=adata.obs['louvain']
    print("OriginalClustering has completed!!!")

def first2prob(adata):
    first2ratio=[name for name in adata.uns.key() if str(name).startswith("prob_matrix")]
    for key_ in first2ratio:
        q_pred=adata.uns[key_]
        q_pred_sort=np.sort(q_pred,axis=1)
        y=q_pred_sort[:,-1]/q_pred_sort[:,-2]
        adata["first2ratio_"+str(key_).split("matrix")[1]]=y

def expand_grid(dictionary):
    from itertools import product
    return pd.DataFrame([row for row in product(*dictionary.values())],columns=dictionary.keys())



from __future__ import division
import os
#import tensorflow as tf
#tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

from . SAE import SAE  # load Stacked autoencoder
from . preprocessing import change_to_continuous
from time import time
import numpy as np
from keras.engine.topology import Layer, InputSpec
from keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau,History
from keras.layers import Dense, Input
from keras.models import Model
from keras.optimizers import SGD
from keras import callbacks
from keras.initializers import VarianceScaling
from sklearn.cluster import KMeans
import scanpy as sc
import pandas as pd
from sklearn.metrics import normalized_mutual_info_score,adjusted_rand_score
import keras.backend as K
from scipy.spatial import distance
from scipy.stats import entropy
import warnings
warnings.filterwarnings('ignore')



class ClusteringLayer(Layer): # Re-define lot of build in functions for Keras
    """
    Clustering layer converts input sample (feature) to soft label, i.e. a vector that represents the probability of the
    sample belonging to each cluster. The probability is calculated with student's t-distribution.

    # Example
    ```
        model.add(ClusteringLayer(n_clusters=10))
    ```
    # Arguments
        n_clusters: number of clusters.
        weights: list of Numpy array with shape `(n_clusters, n_features)` witch represents the initial cluster centers.
        alpha: parameter in Student's t-distribution. Default to 1.0.
    # Input shape
        2D tensor with shape: `(n_samples, n_features)`.
    # Output shape
        2D tensor with shape: `(n_samples, n_clusters)`.
    """

    def __init__(self, n_clusters, weights=None, alpha=1.0, **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(ClusteringLayer, self).__init__(**kwargs)
        self.n_clusters = n_clusters
        self.alpha = alpha
        self.initial_weights = weights
        self.input_spec = InputSpec(ndim=2)

    def build(self, input_shape):
        assert len(input_shape) == 2
        input_dim = input_shape[1]
        self.input_spec = InputSpec(dtype=K.floatx(), shape=(None, input_dim))
        self.clusters = self.add_weight((self.n_clusters, input_dim), initializer='glorot_uniform', name='clustering')
        if self.initial_weights is not None:
            self.set_weights(self.initial_weights)
            del self.initial_weights
        self.built = True

    def call(self, inputs, **kwargs): # The activation function for clustering layer
        """ student t-distribution, as same as used in t-SNE algorithm.
                 q_ij = 1/(1+dist(x_i, u_j)^2), then normalize it.
        Arguments:
            inputs: the variable containing data, shape=(n_samples, n_features)
        Return:
            q: student's t-distribution, or soft labels for each sample. shape=(n_samples, n_clusters)
        """
        q = 1.0 / (1.0 + (K.sum(K.square(K.expand_dims(inputs, axis=1) - self.clusters), axis=2) / self.alpha))
        q **= (self.alpha + 1.0) / 2.0
        q = K.transpose(K.transpose(q) / K.sum(q, axis=1))
        return q

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 2
        return input_shape[0], self.n_clusters

    def get_config(self):
        config = {'n_clusters': self.n_clusters}
        base_config = super(ClusteringLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

class DEC(object):
    def __init__(self,
                 dims,
                 x_all,
                 x_train, # input matrix, row sample, col predictors 
                 y=None, # if provided will trained with supervised
                 alpha=1.0,
                 init='glorot_uniform', #initialization method
                 n_clusters=None,     # Number of Clusters, if provided, the clusters center will be initialized by K-means,
                 louvain_resolution=1.0, # resolution for louvain 
                 n_neighbors=10,    # the 
                 pretrain_epochs=200, # epoch for autoencoder
                 ae_weights=None, #ae_
                 actinlayer1="tanh",# activation for the last layer in encoder, and first layer in the decoder 
                 is_stacked=True,
                 transfer_feature=None,
                 model_weights=None,
                 y_trans=None,
                 softmax=False,
                 ):

        super(DEC, self).__init__()
        self.dims = dims
        self.x_all=x_all #feature n*p, n:number of cells, p: number of genes
        self.x_train = x_train
        self.y=y # for supervised 
        self.y_trans=y_trans
        self.input_dim = dims[0]
        self.n_stacks = len(self.dims) - 1
        self.is_stacked=is_stacked
        self.resolution=louvain_resolution
        self.alpha = alpha
        self.actinlayer1=actinlayer1
        self.transfer_feature=transfer_feature
        self.model_weights=model_weights
        self.softmax=softmax
        self.pretrain_epochs=pretrain_epochs
        if  self.transfer_feature is None:
            self.pretrain(n_neighbors=n_neighbors,epochs=self.pretrain_epochs,n_clusters=n_clusters)
        else:
            self.pretrain_transfer(n_neighbors=n_neighbors,model_weights=self.model_weights,features=transfer_feature,epochs=self.pretrain_epochs,n_clusters=n_clusters,y_trans=self.y_trans)

    def pretrain(self, optimizer='adam', epochs=200, n_neighbors=10,batch_size=256,n_clusters=None):
        #print("Doing DEC: pretrain")  
        sae=SAE(dims=self.dims,drop_rate=0.2,batch_size=batch_size,actinlayer1=self.actinlayer1)# batch_size
        #print('...Pretraining source network...')
        # begin pretraining
        t0 = time()
        if self.is_stacked:
            sae.fit(self.x_all,epochs=epochs)
        else:
            sae.fit2(self.x_all,epochs=epochs)

        self.autoencoder=sae.autoencoders
        self.encoder=sae.encoder
        #print('   ...Pretraining time: ', time() - t0, 'seconds...')
        self.pretrained = True

        #build dec model and initialize model
        features=self.extract_features(self.x_train)
        features=np.asarray(features)
        if self.y is None: # Train data not labeled
            if isinstance(n_clusters,int): # Number of clusters known, use k-means
                print("...number of clusters have been specified, Initializing Cluster centroid  using K-Means")
                kmeans = KMeans(n_clusters=n_clusters, n_init=20)
                Y_pred_init = kmeans.fit_predict(features)
                self.init_pred= np.copy(Y_pred_init)
                self.n_clusters=n_clusters
                cluster_centers=kmeans.cluster_centers_
                self.init_centroid=cluster_centers
            else: # Number of clustered unknow, use unsupervised method
                print("...number of clusters does not know, Initialize Cluster centroid using louvain")
                adata=sc.AnnData(features)
                sc.pp.neighbors(adata, n_neighbors=n_neighbors)
                sc.tl.louvain(adata,resolution=self.resolution)
                Y_pred_init=adata.obs['louvain']
                self.init_pred=np.asarray(Y_pred_init,dtype=int)
                features=pd.DataFrame(features,index=np.arange(0,features.shape[0]))
                Group=pd.Series(self.init_pred,index=np.arange(0,features.shape[0]),name="Group")
                Mergefeature=pd.concat([features,Group],axis=1)
                cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
                self.n_clusters=cluster_centers.shape[0]
                self.init_centroid=cluster_centers
            print("The shape of cluster_centers",cluster_centers.shape)
        else: #  train data is labeled
            #print("y known, initilize Cluster centroid using y")
            # build dec model
            features=pd.DataFrame(features,index=np.arange(0,features.shape[0]))
            Group=pd.Series(self.y.values,index=np.arange(0,features.shape[0]),name="Group")
            Mergefeature=pd.concat([features,Group],axis=1)
            cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
            self.n_clusters=cluster_centers.shape[0]
            self.init_centroid=cluster_centers
            #print("The shape of cluster_center is",cluster_centers.shape)
        if not self.softmax:       # Use dec method to do clustering
            clustering_layer = ClusteringLayer(self.n_clusters, name='clustering')(self.encoder.output)
        else:        # Use softmax to do clustering
            clustering_layer=Dense(self.n_clusters,kernel_initializer="glorot_uniform",name="clustering",activation='softmax')(self.encoder.output)
        self.model = Model(inputs=self.encoder.input, outputs=clustering_layer)



    def pretrain_transfer(self,features,model_weights,y_trans=None,optmizer="adam",n_neighbors=10,epochs=200,batch_size=32,n_clusters=None):
        #y_trans  is a numpy array
        #print("Doing DEC: pretrain_transfer")
        if isinstance(n_clusters,int):
            print("...number of clusters have been specified, Initializing Cluster centroid  using K-Means")
            kmeans = KMeans(n_clusters=n_clusters, n_init=20)
            Y_pred_init = kmeans.fit_predict(features)
            self.init_pred= np.copy(Y_pred_init)
            self.n_clusters=n_clusters
            cluster_centers=kmeans.cluster_centers_
            self.init_centroid=[cluster_centers]
        else:
            #print("The shape of features is",features.shape)
            if y_trans is not None and y_trans.shape[0]==features.shape[0]:
                #print("The shape of y_trans is",y_trans.shape)
                #print("...predicted y_test known, use it to get n_cliusters and init_centroid")
                self.init_pred=y_trans
                features=pd.DataFrame(features,index=np.arange(0,features.shape[0]))
                Group=pd.Series(y_trans,index=np.arange(0,features.shape[0]),name="Group")
                Mergefeature=pd.concat([features,Group],axis=1)
                cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
                self.n_clusters=cluster_centers.shape[0]
                self.init_centroid=cluster_centers
            else:
                print("...number of clusters does not know, Initialize Cluster centroid using louvain")
                #can be replaced by other clustering methods
                adata=sc.AnnData(features)
                sc.pp.neighbors(adata, n_neighbors=n_neighbors) #louvain step1
                sc.tl.louvain(adata,resolution=self.resolution) #louvain step2
                Y_pred_init=adata.obs['louvain']
                self.init_pred=np.asarray(Y_pred_init,dtype=int)
                features=pd.DataFrame(features,index=np.arange(0,features.shape[0]))
                Group=pd.Series(self.init_pred,index=np.arange(0,features.shape[0]),name="Group")
                Mergefeature=pd.concat([features,Group],axis=1)
                cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
                self.n_clusters=cluster_centers.shape[0]
                self.init_centroid=cluster_centers
                print("The shape of cluster_centers",cluster_centers.shape[0])

        sae=SAE(dims=self.dims,drop_rate=0.2,batch_size=batch_size,actinlayer1=self.actinlayer1)# batch_size
        self.autoencoder=sae.autoencoders
        self.encoder=sae.encoder
        clustering_layer=ClusteringLayer(self.n_clusters, name='clustering')(self.encoder.output) # use dec to do clustering
        self.model=Model(self.encoder.input,outputs=clustering_layer)
        #print("The length layers  of self.model",len(self.model.layers))
        for i in range(len(self.model.layers)-2):
            self.model.layers[i+1].set_weights(model_weights[i+1])
        self.model.get_layer("clustering").set_weights([self.init_centroid])
        #fine tunning
    
    def load_weights(self, weights):  # load weights of DEC model
        self.model.load_weights(weights)

    def extract_features(self, x):
        return self.encoder.predict(x)

    def predict(self, x):  # predict cluster labels using the output of clustering layer
        q = self.model.predict(x, verbose=0)
        return q.argmax(1)

    @staticmethod
    def target_distribution(q):
        weight = q ** 2 / q.sum(0)
        return (weight.T / weight.sum(1)).T

    def compile(self, optimizer='sgd', loss='kld'):
        self.model.compile(optimizer=optimizer, loss=loss)
    
    def fit(self,x, maxiter=2e3, epochs_fit=10,batch_size=256, tol=1e-3): # unsupervised
        print("Doing DEC: fit")
        #step1 initial weights by louvain,or Kmeans
        self.model.get_layer(name='clustering').set_weights([self.init_centroid])
        y_pred_last = np.copy(self.init_pred)
        # Step 2: deep clustering
        # logging file
        #y_pred_last=self.init_pred
        loss = 0
        index = 0
        index_array = np.arange(x.shape[0])
        for ite in range(int(maxiter)):
            q = self.model.predict(x, verbose=0)
            p = self.target_distribution(q)  # update the auxiliary target distribution p
            # evaluate the clustering performance
            y_pred = q.argmax(1)

             # check stop criterion
            delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / y_pred.shape[0]
            y_pred_last = np.copy(y_pred)
            if ite > 0 and delta_label < tol:
                print('delta_label ', delta_label, '< tol ', tol)
                print('Reached tolerance threshold. Stopped training.')
                break
            print("The value of delta_label of current",str(ite+1),"th iteration is",delta_label,">= tol",tol)
            #train on whole dataset on prespecified batch_size
            callbacks=[EarlyStopping(monitor='loss',min_delta=10e-4,patience=4,verbose=0,mode='auto')]
            self.model.fit(x=x,y=p,epochs=epochs_fit,batch_size=batch_size,callbacks=callbacks,shuffle=True,verbose=False)

        y0=pd.Series(y_pred)
        print("The final prediction cluster is:")
        print(y0.value_counts())
        Embeded_z=self.encoder.predict(x)
        return Embeded_z,q

    #Show the trajectory of the centroid during iterations
    def fit_trajectory(self,x, maxiter=2e3, epochs_fit=10,batch_size=256, tol=1e-2, celltypes = None, threshold=1/30): # unsupervised
        #print("Doing DEC: fit_trajectory")
        #step1 initial weights by louvain,or Kmeans
        self.model.get_layer(name='clustering').set_weights([self.init_centroid])
        y_pred_last = np.copy(self.init_pred)
        # Step 2: deep clustering
        # logging file
        #y_pred_last=self.init_pred
        loss = 0
        index = 0
        index_array = np.arange(x.shape[0])
        trajectory_z=[] #trajectory embedding
        trajectory_l=[] #trajectory label
        js = []
        centroids_first = self.model.layers[-1].get_weights()
        centroids_diff_all = []
        for i in range(len(centroids_first[0])-1):
            for j in range(i+1, len(centroids_first[0])):
                centroids_diff_all.append(np.sqrt(((centroids_first[0][i]-centroids_first[0][j])**2).sum()))
        print('centroids_diff_all', centroids_diff_all)
        print(len(centroids_diff_all))
        print(self.init_centroid)
#        print(centroids_first)
#        self.model.layers[-1].trainable = False
#        print(self.model.summary())
#        print(self.model.layers[-1].trainable == False)
        weights = self.model.get_weights()
        for ite in range(int(maxiter)):
            old_weights = weights.copy()
            weights = self.model.get_weights()
#            print(weights)
#            print(self.model.layers[-1].trainable == False)
            centroids = self.model.layers[-1].get_weights()
#            print(centroids)
            q = self.model.predict(x, verbose=0)
#            for i in range(len(q)):
#                if sum(q[i]>threshold)==0:
#                    continue
#                for j in range(len(q[i])):
#                    #if q[i][j]<0.1:
#                    if q[i][j]<threshold:
#                        q[i][j]=0
#                q[i] = q[i]/q[i].sum() 
            p = self.target_distribution(q)  # update the auxiliary target distribution p
            # evaluate the clustering performance
           #kl = np.array([[np.where(p[i]!=0, p[i]*np.log(p[i]/q[i]),0) for i in range(len(p))][j].sum() for j in range(len(p))]).sum()
           #print(kl)
           # print(entropy(p,q).sum())
            #print(q.shape)
            #q = pd.DataFrame(q)
            #q.columns = list(celltypes)
            y_pred = q.argmax(1)
            #celltypes = list(np.sort(np.unique(y_pred)))
            celltypes = [celltypes[i] for i in list(np.sort(np.unique(y_pred)))]
            #print(celltypes)
            # check stop criterion
            #delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / y_pred.shape[0]
            #y_pred_last = np.copy(y_pred)
            #if ite > 0 and delta_label < tol:
            #    print('delta_label ', delta_label, '< tol ', tol)
            #    print('Reached tolerance threshold. Stopped training.')
            #    break
            #print("The value of delta_label of current",str(ite+1),"th iteration is",delta_label,">= tol",0.01)
            ##train on whole dataset on prespecified batch_size           
            if ite == 0:
                q_last = np.copy(q)
                js_last = 1000000
                callbacks=[EarlyStopping(monitor='loss',min_delta=10e-4,patience=4,verbose=0,mode='auto')]
                self.model.fit(x=x,y=p,epochs=epochs_fit,batch_size=batch_size,callbacks=callbacks,shuffle=True,verbose=False)
            #if ite < 10:
            #    js.append(distance.jensenshannon(q_last, q).sum())
            #    q_last = np.copy(q)
                #print(js)
            #    callbacks=[EarlyStopping(monitor='loss',min_delta=10e-4,patience=4,verbose=0,mode='auto')]
            #    self.model.fit(x=x,y=p,epochs=epochs_fit,batch_size=batch_size,callbacks=callbacks,shuffle=True,verbose=False)
            #    continue
            if ite>0:
                centroids_diff = [np.sqrt(((centroids[0][i]-centroids_first[0][i])**2).sum()) for i in range(len(centroids[0]))]
                print('centroids_diff: ', centroids_diff)
                #js.append(distance.jensenshannon(q_last, q).sum())
                js = distance.jensenshannon(q_last, q).sum()
                delta_js = js_last-js
                q_last = np.copy(q)
                #print(js_last)
                #print(js)
                js_last = np.copy(js)
                #print(js[ite-10:ite-5])
                #print(js[ite-5:])
                #delta_js = np.mean(js[ite-10:ite-5]) - np.mean(js[ite-5:])
                #delta_js = js[ite-1]-js[ite-2]
                #if delta_js < 0.001 and delta_js>0 and np.mean(js[ite-2:])<np.mean(js[0:3]):
                #if delta_js < 0.01 and delta_js>0 and js[ite-1]<js[0]:
                #if delta_js < 0.01 and delta_js>0 and np.mean(js[ite-5:])< np.mean(js[0:5]):
                #if delta_js < tol and delta_js>0 and np.mean(js[ite-5:])< np.mean(js[0:5]):
                #if delta_js < tol and delta_js>0:
                if ite>1:
                    if sum(np.array(centroids_diff)>1)>0:
                        #print('weights:', weights)
                        #print('old_weights:', old_weights)
                        #print(self.encoder.get_weights())
                        self.model.set_weights(old_weights)
                        q = self.model.predict(x, verbose=0)
#                        print(self.model.get_weights())
                        #print(self.encoder.get_weights())
                        print('Iteration ',ite,': |JSD(Q',ite-2,'||Q',ite-1,')-JSD(Q',ite-1,'||Q',ite,')|=',abs(delta_js),'<', str(tol[0]), sep='')
                        print('Reached tolerance threshold. Stopped training.')
                        break
                    #print("The value of delta_js of current",str(ite+1),"th iteration is",delta_js,">= tol",tol)
                    if ite<=10:
                        print('Iteration ',ite,': |JSD(Q',ite-2,'||Q',ite-1,')-JSD(Q',ite-1,'||Q',ite,')|=',abs(delta_js), sep='')
                    else:
                        print('Iteration ',ite,': |JSD(Q',ite-2,'||Q',ite-1,')-JSD(Q',ite-1,'||Q',ite,')|=',abs(delta_js),'>=',str(tol[0]), sep='')
                callbacks=[EarlyStopping(monitor='loss',min_delta=10e-4,patience=4,verbose=0,mode='auto')]
                self.model.fit(x=x,y=p,epochs=epochs_fit,batch_size=batch_size,callbacks=callbacks,shuffle=True,verbose=False)

            if ite % 10 ==0:
                #print("This is the iteration of ", ite)
                Embeded_z=self.encoder.predict(x) # feature
                q_tmp=self.model.predict(x,verbose=0) # predicted clustering results
                l_tmp=change_to_continuous(q_tmp)
                trajectory_z.append(Embeded_z)
                trajectory_l.append(l_tmp)
        
        #y0=pd.Series(y_pred)
        #print("The final prediction cluster is:")
        #print(y0.value_counts())
        Embeded_z=self.encoder.predict(x)
        return trajectory_z, trajectory_l,  Embeded_z, q, celltypes
        #return ret


    def fit_supervise(self,x,y,epochs=2e3,batch_size=256, celltypes = None):
        #y is 1-D array, Series, or a list, len(y)==x.shape[0]
        print("...Optimizing source network through cell clustering...")
        if self.softmax==False: # Only DEC clustering can set init_centroid
            self.model.get_layer(name='clustering').set_weights([self.init_centroid])
        y0=pd.Series(y,dtype="category") #avoding y is string
        y0=y0.cat.rename_categories(range(len(y0.cat.categories)))
        y_true=pd.get_dummies(pd.Series(y0)).values# coded according to 0,1,...,3
        y_true=y_true+0.00005*np.random.random(y_true.shape)+0.00001 # add some disturb
        y_true=y_true/y_true.sum(axis=1)[:,None]
        callbacks=[EarlyStopping(monitor='loss',min_delta=10e-4,patience=4,verbose=0,mode='auto')]
        self.model.fit(x=x,y=y_true,epochs=int(epochs),batch_size=batch_size,callbacks=callbacks,shuffle=True,verbose=False,validation_split=0.25)
        Embeded_z=self.encoder.predict(x) # feature
        q=self.model.predict(x,verbose=0) # predicted clustering results
        #q = pd.DataFrame(q)
        #q.columns = list(celltypes)
        y_pred = q.argmax(1)
        celltypes = [celltypes[i] for i in list(np.sort(np.unique(y_pred)))]
        #print(celltypes)
        #print(q.shape)

        #return y0, representing the mapping reference for y
        return Embeded_z,q,celltypes



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SpaDecon", 
    version="1.1.1",
    author="Kyle Coleman",
    author_email="kylecole@pennmedicine.upenn.edu",
    description="SpaDecon: Integrating gene expression and histology information to perform cell type deconvolution on spatial transcriptomics data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kylepcoleman87/SpaDecon",
    packages=setuptools.find_packages(),
    install_requires=["keras==2.2.4","pandas==1.2.4","numpy==1.20.1","scipy==1.6.2","scanpy==1.7.0","anndata==0.7.6","sklearn","tensorflow==1.14.0"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)

